import json
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from cryptography.fernet import Fernet, InvalidToken

from .models import AuthenticationResponse


logger = logging.getLogger(__name__)


class AuthenticationCacheABC(ABC):
    """Abstract base class for authentication response caching."""

    @staticmethod
    @abstractmethod
    def save_authentication_response(response: AuthenticationResponse) -> None:
        raise NotImplementedError("Implement Authorization response caching!")

    @staticmethod
    @abstractmethod
    def load_authentication_response() -> Optional[AuthenticationResponse]:
        raise NotImplementedError("Implement Authorization response loading!")


class DurableCache(AuthenticationCacheABC):
    """Local storage for tokens using a file, with optional encryption."""

    def __init__(self, path: str, encryption_key: Optional[bytes] = None) -> None:
        """Initializes token storage with a file path and an optional encryption key.
        :param path: Path to the file where tokens will be stored including the file name and extension.
        :param encryption_key: Optional encryption key for securing tokens. Encryption is disabled if not provided.
        """
        self.path = Path(path)
        self.fernet = Fernet(encryption_key) if encryption_key else None
        logger.info(
            "Authentication cache initialized. Encryption %s.",
            "enabled" if self.fernet else "disabled",
        )

    def save_authentication_response(self, response: AuthenticationResponse) -> None:
        """Save the authentication response to the cache.
        :param response: The authentication response to be cached.
        """
        data = response.model_dump()
        try:
            data_bytes = json.dumps(data).encode()
            if self.fernet:
                encrypted_data = self.fernet.encrypt(data_bytes)
                self.path.write_bytes(encrypted_data)
            else:
                self.path.write_text(json.dumps(data, indent=4))
        except PermissionError:
            logger.error(
                "Permission denied: Unable to cache authentication response to %s",
                str(self.path.absolute()),
            )
        except TypeError as e:
            logger.error("TypeError encountered during caching: %s", str(e))
        except OSError as e:
            logger.error(
                "Failed to cache authentication response due to an OS error: %s", str(e)
            )

    def load_authentication_response(self) -> Optional[AuthenticationResponse]:
        """Load the authentication response from the cache if it exists.
        :return: The authentication response if it exists, otherwise None."""
        if not self.path.exists():
            logger.warning("Authentication cache file not found: %s", self.path)
            return None
        try:
            if self.fernet:
                encrypted_data = self.path.read_bytes()
                decrypted_data = self.fernet.decrypt(encrypted_data).decode()
                data = json.loads(decrypted_data)
            else:
                data = json.loads(self.path.read_text())
            return AuthenticationResponse.model_validate(data)
        except PermissionError:
            logger.error(
                "Permission denied: Unable to load authentication response from %s",
                self.path,
            )
            return None
        except (json.JSONDecodeError, InvalidToken) as e:
            logger.error("Failed to load authentication response: %s", str(e))
            return None
        except OSError as e:
            logger.error(
                "Failed to load authentication response due to an OS error: %s", str(e)
            )
            return None


class InMemoryCache(AuthenticationCacheABC):
    """In-memory storage for authentication response."""

    def __init__(self) -> None:
        self.response = None
        logger.info("In-memory cache initialized.")

    def save_authentication_response(self, response: AuthenticationResponse) -> None:
        """Save the authentication response to the cache if it exists.
        :param response: The authentication response to be cached."""
        self.response = response

    def load_authentication_response(self) -> Optional[AuthenticationResponse]:
        """Load the authentication response from memory if it exists.
        :return: The authentication response if it exists, otherwise None."""
        if self.response:
            return self.response
        else:
            return None



from decimal import Decimal
from typing import List, Any, Literal, Optional

from pydantic import (
    BaseModel,
    constr,
    condecimal,
    conint,
    model_validator,
    field_serializer,
)
from ..models import InstrumentType, Direction


MarketStatusType = Literal[
    "TRADEABLE",
    "CLOSED",
    "EDITS_ONLY",
    "OFFLINE",
    "ON_AUCTION",
    "ON_AUCTION_NO_EDITS",
    "SUSPENDED",
]

OrderType = Literal["LIMIT", "MARKET", "QUOTE"]

TimeInForce = Literal["EXECUTE_AND_ELIMINATE", "FILL_OR_KILL"]


class Market(BaseModel):
    instrumentName: str
    expiry: str
    epic: str
    instrumentType: InstrumentType
    lotSize: float
    high: float
    low: float
    percentageChange: float
    netChange: float
    bid: float
    offer: float
    updateTime: str
    updateTimeUTC: str
    delayTime: int
    streamingPricesAvailable: bool
    marketStatus: MarketStatusType
    scalingFactor: int


class Position(BaseModel):
    contractSize: condecimal(decimal_places=2)
    controlledRisk: bool
    createdDate: str
    createdDateUTC: str
    currency: str
    dealId: str
    dealReference: str
    direction: Direction
    level: condecimal(decimal_places=2)
    limitLevel: Optional[condecimal(decimal_places=2)] = None
    limitedRiskPremium: Optional[condecimal(decimal_places=2)] = None
    size: condecimal(decimal_places=2)
    stopLevel: Optional[condecimal(decimal_places=2)] = None
    trailingStep: Optional[conint(ge=0)] = None
    trailingStopDistance: Optional[condecimal(decimal_places=2)] = None


class OpenPosition(BaseModel):
    position: Position
    market: Market


class OpenPositions(BaseModel):
    positions: List[OpenPosition]


class CreatePosition(BaseModel):
    currencyCode: constr(pattern=r"^[A-Z]{3}$")
    direction: Direction
    epic: constr(pattern=r"^[A-Za-z0-9._]{6,30}$")
    expiry: constr(pattern=r"^(\d{2}-)?[A-Z]{3}-\d{2}|-|DFB$")
    forceOpen: bool
    guaranteedStop: bool
    orderType: OrderType
    timeInForce: TimeInForce
    trailingStop: bool
    dealReference: Optional[constr(pattern=r"^[A-Za-z0-9_\-.]{1,30}$")] = None
    level: Optional[condecimal(decimal_places=12)] = None
    limitDistance: Optional[condecimal(decimal_places=2)] = None
    limitLevel: Optional[condecimal(decimal_places=2)] = None
    quoteId: Optional[constr(pattern=r"^[A-Za-z0-9]+$")] = None
    size: condecimal(decimal_places=2, gt=0)
    stopDistance: Optional[condecimal(decimal_places=2)] = None
    stopLevel: Optional[condecimal(decimal_places=2)] = None
    trailingStopIncrement: Optional[condecimal(decimal_places=2)] = None

    @field_serializer(
        "level",
        "limitDistance",
        "limitLevel",
        "size",
        "stopDistance",
        "stopLevel",
        "trailingStopIncrement",
        mode="plain",
    )
    def serialize_decimal(self, v: Optional[Decimal], _info) -> float:
        if v is not None:
            return float(v)

    @model_validator(mode="before")
    @classmethod
    def check_unique_constraints(cls, data: Any):
        if data.get("limitLevel") is not None and data.get("limitDistance") is not None:
            raise ValueError("Set only one of limitLevel or limitDistance.")
        if data.get("stopLevel") is not None and data.get("stopDistance") is not None:
            raise ValueError("Set only one of stopLevel or stopDistance.")
        return data

    @model_validator(mode="before")
    @classmethod
    def check_force_open_constraints(cls, data: Any):
        if not data.get("forceOpen") and any(
            [
                data.get("limitDistance") is not None,
                data.get("limitLevel") is not None,
                data.get("stopDistance") is not None,
                data.get("stopLevel") is not None,
            ]
        ):
            raise ValueError(
                "forceOpen must be true if limit or stop constraints are set."
            )
        return data

    @model_validator(mode="before")
    @classmethod
    def check_guaranteed_stop_constraints(cls, data: Any):
        if data.get("guaranteedStop") and not (
            bool(data.get("stopLevel")) ^ bool(data.get("stopDistance"))
        ):
            raise ValueError(
                "When guaranteedStop is true, specify exactly one of stopLevel or stopDistance."
            )
        return data

    @model_validator(mode="before")
    @classmethod
    def check_order_type_constraints(cls, data: Any):
        order_type = data.get("orderType")
        if order_type == "LIMIT":
            if data.get("quoteId") is not None:
                raise ValueError("Do not set quoteId when orderType is LIMIT.")
            if data.get("level") is None:
                raise ValueError("Set level when orderType is LIMIT.")
        elif order_type == "MARKET":
            if any([data.get("level") is not None, data.get("quoteId") is not None]):
                raise ValueError(
                    "Do not set level or quoteId when orderType is MARKET."
                )
        elif order_type == "QUOTE":
            if not all(
                [data.get("level") is not None, data.get("quoteId") is not None]
            ):
                raise ValueError("Set both level and quoteId when orderType is QUOTE.")
        return data

    @model_validator(mode="before")
    @classmethod
    def check_trailing_stop_constraints(cls, data: Any):
        if data.get("trailingStop"):
            if data.get("stopLevel") is not None:
                raise ValueError("Do not set stopLevel when trailingStop is true.")
            if data.get("guaranteedStop"):
                raise ValueError(
                    "guaranteedStop must be false when trailingStop is true."
                )
            if not all(
                [
                    data.get("stopDistance") is not None,
                    data.get("trailingStopIncrement") is not None,
                ]
            ):
                raise ValueError(
                    "Set both stopDistance and trailingStopIncrement when trailingStop is true."
                )
        return data


class ClosePosition(BaseModel):
    direction: Direction
    orderType: OrderType
    size: condecimal(gt=0, decimal_places=2)
    timeInForce: TimeInForce
    quoteId: Optional[str] = None
    dealId: Optional[constr(pattern=".{1,30}")] = None
    epic: Optional[constr(pattern="[A-Za-z0-9._]{6,30}")] = None
    expiry: Optional[constr(pattern="(\\d{2}-)?[A-Z]{3}-\\d{2}|-|DFB")] = None
    level: Optional[condecimal(decimal_places=12)] = None

    @field_serializer("level", "size", mode="plain")
    def serialize_decimal(self, v: Optional[Decimal], _info) -> Optional[float]:
        if v is not None:
            return float(v)

    @model_validator(mode="before")
    @classmethod
    def check_order_type(cls, data: Any):
        if data.get("orderType") == "QUOTE" and (
            data.get("quoteId") is None or data.get("level") is None
        ):
            raise ValueError("quoteId is required when orderType is QUOTE.")

        if (
            data.get("orderType") == "MARKET"
            and (data.get("level") is not None or data.get("quoteId")) is not None
        ):
            raise ValueError(
                "level and quoteId are not allowed when orderType is MARKET."
            )

        if data.get("orderType") == "LIMIT" and data.get("quoteId") is not None:
            raise ValueError("quoteId is not allowed when orderType is LIMIT.")

        if data.get("orderType") == "LIMIT" and data.get("level") is None:
            raise ValueError("level is required when orderType is LIMIT.")

        return data

    @model_validator(mode="before")
    @classmethod
    def check_unique_constraints(cls, data: Any):
        if data.get("dealId") is not None and data.get("epic") is not None:
            raise ValueError("Set only one of dealId or epic.")
        if data.get("dealId") is None and data.get("epic") is None:
            raise ValueError("Set one of dealId or epic.")
        return data

    @classmethod
    def from_create(cls, position: CreatePosition):
        """Create a ClosePosition from a CreatePosition.
        :param position: CreatePosition. The position to close.
        :return: ClosePosition. The position to close.
        """
        return cls(
            direction="BUY" if position.direction == "SELL" else "SELL",
            orderType=position.orderType,
            size=position.size,
            timeInForce=position.timeInForce,
            epic=position.epic,
            expiry=position.expiry,
            level=position.level,
        )


class UpdatePosition(BaseModel):
    guaranteedStop: Optional[bool] = None
    limitLevel: Optional[condecimal(decimal_places=2)] = None
    stopLevel: Optional[condecimal(decimal_places=2)] = None
    trailingStop: Optional[bool] = None
    trailingStopDistance: Optional[condecimal(decimal_places=2)] = None
    trailingStopIncrement: Optional[condecimal(decimal_places=2)] = None

    @field_serializer(
        "limitLevel",
        "stopLevel",
        "trailingStopDistance",
        "trailingStopIncrement",
        mode="plain",
    )
    @classmethod
    def serialize_decimal(cls, v: Optional[Decimal], _info) -> Optional[float]:
        if v is not None:
            return float(v)

    @model_validator(mode="before")
    @classmethod
    def validate_trailing_stop_constraints(cls, data: dict):
        if data.get("trailingStop"):
            cls._validate_trailing_stop_true(data)
        elif data.get("trailingStop") == False:
            cls._validate_trailing_stop_false(data)
        return data

    @model_validator(mode="before")
    @classmethod
    def validate_guaranteed_stop_constraints(cls, data: dict):
        if data.get("guaranteedStop"):
            cls._validate_guaranteed_stop_true(data)
        return data

    @classmethod
    def _validate_trailing_stop_true(cls, data: dict):
        if data.get("guaranteedStop"):
            raise ValueError(
                "If trailingStop is true, then guaranteedStop must be false."
            )
        if any(
            data.get(field) is None
            for field in ["trailingStopDistance", "trailingStopIncrement", "stopLevel"]
        ):
            raise ValueError(
                "If trailingStop is true, then trailingStopDistance, trailingStopIncrement, and stopLevel must be set."
            )

    @classmethod
    def _validate_trailing_stop_false(cls, data: dict):
        if any(
            data.get(field) is not None
            for field in ["trailingStopDistance", "trailingStopIncrement"]
        ):
            raise ValueError(
                "If trailingStop is false, then DO NOT set trailingStopDistance or trailingStopIncrement."
            )

    @classmethod
    def _validate_guaranteed_stop_true(cls, data: dict):
        if data.get("stopLevel") is None:
            raise ValueError("If guaranteedStop is true, then stopLevel must be set.")
        if data.get("trailingStop"):
            raise ValueError("guaranteedStop and trailingStop cannot both be true.")



from datetime import datetime
from decimal import Decimal
from typing import Optional, List, Literal, Any

from pydantic import (
    BaseModel,
    condecimal,
    field_validator,
    constr,
    model_validator,
    field_serializer,
)
from ..models import InstrumentType, Direction


OrderType = Literal["LIMIT", "STOP"]


TimeInForce = Literal["GOOD_TILL_CANCELLED", "GOOD_TILL_DATE"]


MarketStatus = Literal[
    "CLOSED",
    "EDITS_ONLY",
    "OFFLINE",
    "ON_AUCTION",
    "ON_AUCTION_NO_EDITS",
    "SUSPENDED",
    "TRADEABLE",
]


class MarketData(BaseModel):
    instrumentName: str
    exchangeId: str
    expiry: str
    marketStatus: MarketStatus
    epic: str
    instrumentType: InstrumentType
    lotSize: condecimal(decimal_places=2)
    high: condecimal(decimal_places=2)
    low: condecimal(decimal_places=2)
    percentageChange: condecimal(decimal_places=2)
    netChange: condecimal(decimal_places=2)
    bid: condecimal(decimal_places=2)
    offer: condecimal(decimal_places=2)
    updateTime: str
    updateTimeUTC: str
    delayTime: int
    streamingPricesAvailable: bool
    scalingFactor: int


class WorkingOrderData(BaseModel):
    dealId: str
    direction: Direction
    epic: str
    orderSize: condecimal(decimal_places=2)
    orderLevel: condecimal(decimal_places=2)
    timeInForce: TimeInForce
    goodTillDate: Optional[datetime] = None
    goodTillDateISO: Optional[datetime] = None
    createdDate: datetime
    createdDateUTC: datetime
    guaranteedStop: bool
    orderType: OrderType
    stopDistance: Optional[condecimal(decimal_places=2)] = None
    limitDistance: Optional[condecimal(decimal_places=2)] = None
    currencyCode: str
    dma: Optional[bool] = None
    limitedRiskPremium: Optional[condecimal(decimal_places=2)] = None

    @field_validator("goodTillDate", mode="before")
    @classmethod
    def parse_good_till_date(cls, v: Optional[str]) -> Optional[datetime]:
        if v is not None:
            try:
                return datetime.strptime(v, "%Y/%m/%d %H:%M")
            except ValueError:
                raise ValueError(f"Invalid datetime format for goodTillDate value {v}")

    @field_validator("createdDate", mode="before")
    @classmethod
    def parse_created_date(cls, v: Optional[str]) -> Optional[datetime]:
        try:
            return datetime.strptime(v, "%Y/%m/%d %H:%M:%S:%f")
        except ValueError:
            raise ValueError(f"Invalid datetime format for createdDate value {v}")


class WorkingOrder(BaseModel):
    workingOrderData: WorkingOrderData
    marketData: MarketData


class WorkingOrders(BaseModel):
    workingOrders: List[WorkingOrder]


class CreateWorkingOrder(BaseModel):
    currencyCode: constr(pattern="[A-Z]{3}")
    direction: Direction
    epic: str
    expiry: constr(pattern="(\\d{2}-)?[A-Z]{3}-\\d{2}|-|DFB")
    forceOpen: bool = False
    goodTillDate: constr(pattern="(\\d{4}/\\d{2}/\\d{2} \\d{2}:\\d{2}:\\d{2}|\\d*)")
    level: condecimal(decimal_places=2)
    size: condecimal(decimal_places=12)
    type: OrderType
    guaranteedStop: bool = False
    timeInForce: TimeInForce = "GOOD_TILL_CANCELLED"
    dealReference: Optional[str] = None
    limitDistance: Optional[condecimal(decimal_places=2)] = None
    limitLevel: Optional[condecimal(decimal_places=2)] = None
    stopDistance: Optional[condecimal(decimal_places=2)] = None
    stopLevel: Optional[condecimal(decimal_places=2)] = None

    @field_serializer(
        "level",
        "size",
        "limitDistance",
        "limitLevel",
        "stopDistance",
        "stopLevel",
        mode="plain",
    )
    def serialize_decimal(self, v: Optional[Decimal], _info) -> float:
        if v is not None:
            return float(v)

    @model_validator(mode="before")
    @classmethod
    def check_unique_constraints(cls, data: Any):
        """[Constraint: Set only one of {limitLevel,limitDistance}]
        [Constraint: Set only one of {stopLevel,stopDistance}]"""
        if data.get("limitLevel") is not None and data.get("limitDistance") is not None:
            raise ValueError("Set only one of limitLevel or limitDistance.")
        if data.get("stopLevel") is not None and data.get("stopDistance") is not None:
            raise ValueError("Set only one of stopLevel or stopDistance.")
        return data

    @model_validator(mode="before")
    @classmethod
    def validate_good_till_date(self, data: Any):
        if (
            data.get("timeInForce") == "GOOD_TILL_DATE"
            and data.get("goodTillDate") is None
        ):
            raise ValueError(
                "timeInForce GOOD_TILL_DATE requires a goodTillDate value."
            )
        return data

    @model_validator(mode="before")
    @classmethod
    def check_guaranteed_stop_constraints(cls, data: Any):
        if data.get("guaranteedStop") and not (
            bool(data.get("stopLevel")) ^ bool(data.get("stopDistance"))
        ):
            raise ValueError(
                "When guaranteedStop is true, specify exactly one of stopLevel or stopDistance."
            )
        return data



from collections import namedtuple
from typing import List, Optional

from pydantic import BaseModel


Tokens = namedtuple("Tokens", ["cst_token", "x_security_token"])


class AccountFinancials(BaseModel):
    balance: float
    deposit: float
    profitLoss: float
    available: float


class AccountSummary(BaseModel):
    accountId: str
    accountName: str
    preferred: bool
    accountType: str


class AccountInfo(BaseModel):
    accountType: str
    accountInfo: AccountFinancials
    currencyIsoCode: str
    currencySymbol: str
    currentAccountId: str
    lightstreamerEndpoint: str
    accounts: List[AccountSummary]
    clientId: str
    timezoneOffset: int
    hasActiveDemoAccounts: bool
    hasActiveLiveAccounts: bool
    trailingStopsEnabled: bool
    reroutingEnvironment: Optional[str] = None
    dealingEnabled: bool


class AuthenticationResponse(BaseModel):
    cst_token: str
    x_security_token: str
    expiry: Optional[int] = None
    account_info: Optional[AccountInfo] = None

    @property
    def tokens(self) -> Tokens:
        return Tokens(self.cst_token, self.x_security_token)



from .models import OpenPositions, OpenPosition, CreatePosition, ClosePosition, UpdatePosition
from .service import PositionService, PositionsError


__all__ = [
    "OpenPositions",
    "OpenPosition",
    "CreatePosition",
    "ClosePosition",
    "UpdatePosition",
    "PositionService",
    "PositionsError",
]



from ig_trading_lib.authentication.models import (
    AccountInfo,
    AccountSummary,
    AccountFinancials,
    AuthenticationResponse,
    Tokens,
)

__all__ = [
    "AccountInfo",
    "AccountSummary",
    "AccountFinancials",
    "AuthenticationResponse",
    "Tokens",
]



import logging

import requests
from pydantic import ValidationError

from ig_trading_lib import Tokens
from .models import WorkingOrders, CreateWorkingOrder
from ..models import DealReference


logger = logging.getLogger(__name__)


class OrderException(Exception):
    """Exception raised for errors in the order process."""
    pass


class OrderService:
    def __init__(self, api_key: str, tokens: Tokens, base_url: str):
        self.api_key = api_key
        self.tokens = tokens
        self.base_url = base_url

    @property
    def headers(self) -> dict:
        return {
            "Content-Type": "application/json; charset=utf-8",
            "Accept": "application/json; charset=utf-8",
            "Version": "2",
            "X-IG-API-KEY": self.api_key,
            "X-SECURITY-TOKEN": self.tokens.x_security_token,
            "CST": self.tokens.cst_token,
        }


    def get_working_orders(self) -> WorkingOrders:
        """Get working orders list"""
        url = f"{self.base_url}/gateway/deal/workingorders"
        try:
            response = requests.get(url, headers=self.headers)
            if response.status_code == 200:
                return WorkingOrders.model_validate(response.json())
            else:
                raise OrderException(
                    "Working orders failed with status code %s: %s"
                    % (response.status_code, response.text)
                )
        except ValidationError as e:
            raise OrderException("Invalid working orders response: %s" % e)
        except requests.RequestException as e:
            raise OrderException("Working orders request failed: %s" % e)


    def create_working_order(self, working_order: CreateWorkingOrder) -> DealReference:
        """Create a new working order"""
        url = f"{self.base_url}/gateway/deal/workingorders/otc"
        try:
            response = requests.post(url, headers=self.headers, json=working_order.dict())
            if response.status_code == 200:
                return DealReference.model_validate(response.json())
            else:
                raise OrderException(
                    "Create working order failed with status code %s: %s"
                    % (response.status_code, response.text)
                )
        except ValidationError as e:
            raise OrderException("Invalid create working order response: %s" % e)
        except requests.RequestException as e:
            raise OrderException("Create working order request failed: %s" % e)





from typing import Literal
from pydantic import BaseModel


class DealReference(BaseModel):
    dealReference: str


InstrumentType = Literal[
    "SHARES",
    "BINARY",
    "BUNGEE_CAPPED",
    "BUNGEE_COMMODITIES",
    "BUNGEE_CURRENCIES",
    "BUNGEE_INDICES",
    "COMMODITIES",
    "CURRENCIES",
    "INDICES",
    "KNOCKOUTS_COMMODITIES",
    "KNOCKOUTS_CURRENCIES",
    "KNOCKOUTS_INDICES",
    "KNOCKOUTS_SHARES",
    "OPT_COMMODITIES",
    "OPT_CURRENCIES",
    "OPT_INDICES",
    "OPT_RATES",
    "OPT_SHARES",
    "RATES",
    "SECTORS",
    "SPRINT_MARKET",
    "TEST_MARKET",
    "UNKNOWN",
]


Direction = Literal["BUY", "SELL"]


import logging
from datetime import datetime
from typing import Optional

import requests
from pydantic import ValidationError

from .cache import AuthenticationCacheABC
from .models import AuthenticationResponse


logger = logging.getLogger(__name__)


class AuthenticationError(Exception):
    """Exception raised for errors in the authentication process."""


class AuthenticationService:
    """Handles authentication with the IG Rest API."""

    def __init__(
        self,
        api_key: str,
        account_identifier: str,
        account_password: str,
        base_url: str,
        cache: Optional[AuthenticationCacheABC] = None,
    ):
        """Initialize the authentication service.
        :param api_key: Your IG API key.
        :param account_identifier: Your IG account identifier.
        :param account_password: Your IG account password.
        :param base_url: The base URL for the IG API (live or demo).
        :param cache: An optional cache to store the authentication response. Defaults to no caching
        """
        self.api_key = api_key
        self.account_identifier = account_identifier
        self.account_password = account_password
        self.base_url = base_url
        self.cache = cache
        logger.info(
            "Initialized authentication service. Caching %s.",
            "enabled" if self.cache else "disabled",
        )

    @property
    def url(self) -> str:
        """IG API endpoint for authentication."""
        return f"{self.base_url}/gateway/deal/session"

    @property
    def data(self) -> dict:
        """Your IG account credentials in a dictionary."""
        return {
            "identifier": self.account_identifier,
            "password": self.account_password,
        }

    @property
    def headers(self) -> dict:
        """Setting up the request headers."""
        return {
            "Content-Type": "application/json; charset=utf-8",
            "Accept": "application/json; charset=utf-8",
            "X-IG-API-KEY": self.api_key,
            "Version": "2",
        }

    def __load_from_cache(self) -> Optional[AuthenticationResponse]:
        """Load the authentication response from the cache if it exists and is not expired."""
        if self.cache:
            cached_response = self.cache.load_authentication_response()
            if cached_response and cached_response.expiry > datetime.now().timestamp():
                return cached_response
        return None

    def __save_to_cache(self, response: AuthenticationResponse) -> None:
        """Save the authentication response to the cache if it exists."""
        if self.cache:
            self.cache.save_authentication_response(response)

    def __authenticate(self) -> AuthenticationResponse:
        """Perform the POST request to authenticate with the IG REST API."""
        try:
            response = requests.post(self.url, json=self.data, headers=self.headers)
        except requests.RequestException as e:
            raise AuthenticationError(f"Authentication request failed: %s" % e)

        if response.status_code == 200:
            try:
                expiry = int(response.headers.get("expires", 0)) or int(
                    datetime.now().timestamp() + 6 * 3600
                )
                auth_response = AuthenticationResponse(  # raises ValidationError
                    cst_token=response.headers.get("CST"),
                    x_security_token=response.headers.get("X-SECURITY-TOKEN"),
                    expiry=expiry,
                    account_info=response.json(),  # raises ValueError
                )
                return auth_response
            except ValidationError as e:
                raise AuthenticationError("Invalid authentication response: %s" % e)
            except ValueError as e:
                raise AuthenticationError(
                    "Invalid account info in authentication response: %s" % e
                )
        else:
            raise AuthenticationError(
                "Authentication failed with status code %s: %s"
                % (response.status_code, response.text)
            )

    def __authenticate_and_cache(self) -> AuthenticationResponse:
        """Authenticate with the IG REST API and save the response to the cache."""
        response = self.__authenticate()
        self.__save_to_cache(response)
        return response

    def authenticate(self) -> AuthenticationResponse:
        """Authenticate with the IG REST API."""
        return self.__load_from_cache() or self.__authenticate_and_cache()



import logging

import requests
from pydantic import ValidationError

from ig_trading_lib import Tokens
from .models import (
    OpenPositions,
    OpenPosition,
    CreatePosition,
    ClosePosition,
    UpdatePosition,
)
from ..models import DealReference


logger = logging.getLogger(__name__)


class PositionsError(Exception):
    """Exception raised for errors in the positions process."""


class PositionService:
    def __init__(self, api_key: str, tokens: Tokens, base_url: str):
        """Initialize the position service.
        :param api_key: Your IG API key.
        :param tokens: Authentication tokens for the IG API.
        :param base_url: The base URL for the IG API (live or demo). e.g: https://demo-api.ig.com/gateway/deal
        """
        self.api_key = api_key
        self.tokens = tokens
        self.base_url = base_url

    @property
    def headers(self) -> dict:
        return {
            "Content-Type": "application/json; charset=utf-8",
            "Accept": "application/json; charset=utf-8",
            "Version": "2",
            "X-IG-API-KEY": self.api_key,
            "X-SECURITY-TOKEN": self.tokens.x_security_token,
            "CST": self.tokens.cst_token,
        }

    def get_open_position_by_deal_id(self, deal_id: str) -> OpenPosition:
        """Get open position by deal ID for the authenticated account.
        :param deal_id: str. The deal ID of the open position.
        :return: OpenPosition. The open position for the authenticated account.
        """
        url = f"{self.base_url}/gateway/deal/positions/{deal_id}"
        try:
            response = requests.get(url, headers=self.headers)
            if response.status_code == 200:
                return OpenPosition.model_validate(response.json())
            else:
                raise PositionsError(
                    "Open position request failed with status code %s: %s"
                    % (response.status_code, response.text)
                )
        except ValidationError as e:
            raise PositionsError("Invalid open position response: %s" % e)
        except requests.RequestException as e:
            raise PositionsError("Open position request failed: %s" % e)

    def get_open_positions(self) -> OpenPositions:
        """Get open positions for the authenticated account.
        :return: OpenPositions. The open positions for the authenticated account.
        """
        url = f"{self.base_url}/gateway/deal/positions"
        try:
            response = requests.get(url, headers=self.headers)
            if response.status_code == 200:
                return OpenPositions.model_validate(response.json())
            else:
                raise PositionsError(
                    "Open positions request failed with status code %s: %s"
                    % (response.status_code, response.text)
                )
        except ValidationError as e:
            raise PositionsError("Invalid open positions response: %s" % e)
        except requests.RequestException as e:
            raise PositionsError("Open positions request failed: %s" % e)

    def create_position(self, create: CreatePosition) -> DealReference:
        """Create a new position for the authenticated account.
        :param create: CreatePosition. The position to create.
        :return: DealReference e.g: {'dealReference': 'DIAAAABBBCCC123'}
        """

        url = f"{self.base_url}/gateway/deal/positions/otc"
        try:
            response = requests.post(
                url, headers=self.headers, json=create.model_dump()
            )
            if response.status_code == 200:
                return DealReference.model_validate(response.json())
            else:
                raise PositionsError(
                    "Create position request failed with status code %s: %s"
                    % (response.status_code, response.text)
                )
        except requests.RequestException as e:
            raise PositionsError("Create position request failed: %s" % e)

    def close_position(self, close: ClosePosition) -> DealReference:
        """Close a position for the authenticated account.
        :param close: ClosePosition. The position to close.
        :return: DealReference e.g: {'dealReference': 'DIAAAABBBCCC123'}
        """

        url = f"{self.base_url}/gateway/deal/positions/otc"
        try:
            response = requests.post(url, headers=self.headers, json=close.model_dump())
            if response.status_code == 200:
                return DealReference.model_validate(response.json())
            else:
                raise PositionsError(
                    "Close position request failed with status code %s: %s"
                    % (response.status_code, response.text)
                )
        except requests.RequestException as e:
            raise PositionsError("Close position request failed: %s" % e)

    def update_position(self, deal_id: str, update: UpdatePosition) -> DealReference:
        """Update a position for the authenticated account.
        :param deal_id: str. The deal ID of the position to update.
        :param update: UpdatePosition. Position update details.
        :return: DealReference e.g: {'dealReference': 'DIAAAABBBCCC123'}
        """

        url = f"{self.base_url}/gateway/deal/positions/otc/{deal_id}"
        try:
            response = requests.put(url, headers=self.headers, json=update.model_dump())
            if response.status_code == 200:
                return DealReference.model_validate(response.json())
            else:
                raise PositionsError(
                    "Update position request failed with status code %s: %s"
                    % (response.status_code, response.text)
                )
        except requests.RequestException as e:
            raise PositionsError("Update position request failed: %s" % e)
